template!() { /* proc-macro */ }Expand description
Generates repeated Rust code from one or more table-driven input clauses.
See the crate-level documentation for the input syntax, splice behavior, and examples.
template!() { /* proc-macro */ }Generates repeated Rust code from one or more table-driven input clauses.
See the crate-level documentation for the input syntax, splice behavior, and examples.